/*
 * A process monitor abstraction.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/thread.h>

#include <ipc/cap_alloc.h>
#include <ipc/map.h>
#include <systypes/base.h>

#include <stdio.h>

#include "process_monitor.h"
#include "parent_notification_object_server.h"



/* A process monitor receiving signals from a task. */

ProcessMonitor::ProcessMonitor()
: NotificationSupport()
{
}

ipc_server_default_config_type ProcessMonitor::config()
{
  return config_ParentNotificationObject;
}



/* Close the process monitor. */

void ProcessMonitor::close()
{
  printf("Process monitor closing...\n");
}



/* Capability management. */

void ProcessMonitor::set_pager(l4_cap_idx_t cap, l4_cap_idx_t mapped_cap)
{
  _pager = cap;
  _mapped_pager = mapped_cap;
}

void ProcessMonitor::set_parent(l4_cap_idx_t cap, l4_cap_idx_t mapped_cap)
{
  _parent = cap;
  _mapped_parent = mapped_cap;
}

void ProcessMonitor::set_task(l4_cap_idx_t cap, l4_cap_idx_t mapped_cap)
{
  _task = cap;
  _mapped_task = mapped_cap;
}

void ProcessMonitor::set_thread(l4_cap_idx_t cap, l4_cap_idx_t mapped_cap)
{
  _thread = cap;
  _mapped_thread = mapped_cap;
}



/* Lifecycle management. */

void ProcessMonitor::pager_ended()
{
  /* Release the program thread and its reference to this object, leaving only
     the program initiator with a reference. When the initiator releases its
     reference, this object will be closed. */

  if (l4_is_valid_cap(_thread))
  {
    ipc_unmap_capability(_task, _mapped_thread);
    ipc_cap_free_um(_thread);
    _thread = L4_INVALID_CAP;

    ipc_unmap_capability(_task, _mapped_parent);
  }

  /* Release the program task. This will release any capabilities that might be
     blocking the initiator, such as those involving pipes or streams. */

  if (l4_is_valid_cap(_task))
  {
    ipc_unmap_capability(_task, _mapped_task);
    ipc_cap_free_um(_task);
    _task = L4_INVALID_CAP;
  }
}



/* Receive signals from a task. */

long ProcessMonitor::signal(unsigned long sig, unsigned long val)
{
  notify_all(NOTIFY_TASK_SIGNAL, (notify_values_t) {sig, val});

  /* Handle the termination event. */

  if (sig == 0)
  {
    printf("Signal from task.\n");

    /* Cancel any IPC to avoid spurious pager warnings. */

    l4_thread_ex_regs(_pager, ~0UL, ~0UL, L4_THREAD_EX_REGS_CANCEL);

    /* Once the program exits, the IPC gate connecting the program with its
       internal pager can be released. This will cause the internal pager to
       finish, which is then handled by the external pager. */

    if (l4_is_valid_cap(_pager))
    {
      ipc_cap_free_um(_pager);
      _pager = L4_INVALID_CAP;
    }

    /* The thread is not released here since the initiator of the process
       needs to be able to receive a reference to this object, but if the
       process terminates and the thread is released, this object may be
       released before the initiator obtains a reference to it. */
  }

  return L4_EOK;
}



/* Subscribe to notifications. */

long ProcessMonitor::subscribe(l4_cap_idx_t notifier, notify_flags_t flags)
{
  unsubscribe();

  _notifier = notifier;
  return NotificationSupport::subscribe(_notifier, flags);
}

/* Unsubscribe from notifications. */

long ProcessMonitor::unsubscribe()
{
  if (l4_is_valid_cap(_notifier))
  {
    long err = NotificationSupport::unsubscribe(_notifier);
    _notifier = L4_INVALID_CAP;
    return err;
  }
  else
    return L4_EOK;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
